/*  Routines related to mailing messages about items to users.
 */
#define MAILTO_C

#include "config.h"		/* for definition of MAILPROG */

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <pwd.h>

#include "message.h"
#include "mailto.h"

static StringList *read_stringlist(char *fname);

UserList *userlist_head = NULL;

		/* Add the note to list of notes for user.
		 */
void make_note(char *note, uid_t uid)
{
  UserList *cursor;
  StringList *new_note;
  char *s;

  message(LOG_DEBUG,"make_note(%s,%d)\n",note,uid);

  new_note = malloc(sizeof(StringList));/* Allocate space for string node */
  s = malloc(strlen(note)+1);	/* Allocate space for the string itself */
  if(new_note == NULL || s == NULL) {
    message(LOG_FATAL, "error allocating memory: %s\n", strerror(errno));
  }
  strcpy(s,note);		/* Save the string contents */
  new_note->string = s;		/* Attach the string to the node */
  new_note->next = NULL;	/* New note is always the tail */


				/* Now find out if this user already has
				   some notes.
				 */
  for(cursor = userlist_head; cursor != NULL; cursor = cursor->next) {
    if(cursor->uid == uid) {
      break;
    }
  }

  if( cursor == NULL ) {	/* End of list reached: user not found */

				/* Create new UserList node for this user */
    UserList *newuser = malloc(sizeof(UserList));
    if(newuser == NULL) {
      message(LOG_FATAL, "error allocating memory: %s\n", strerror(errno));
    }

				/* Insert the new user onto front of
				   list, to take advantage of likelihood
				   that the same user will be the
				   subject of the next calls.
				*/
    newuser->uid = uid;
    newuser->note_list = new_note;
    newuser->note_tail = new_note;
    newuser->next = userlist_head;
    userlist_head = newuser;

  }
  else {

				/* Insert new note onto tail of string list.
				 */
    cursor->note_tail->next = new_note;
    cursor->note_tail = new_note;
  }
}


void mail_notices(char *subject, char *preamble_filename)
{
  UserList *cursor;
  StringList *preamble=read_stringlist(preamble_filename);




  for(cursor = userlist_head; cursor != NULL; cursor = cursor->next) {
				/* Look up the user's name from user id */
    struct passwd *pwd=getpwuid(cursor->uid);
    char *username;
    if( pwd != NULL ) {
      if( at_host != NULL) {	/* Tack on "@host.domain" after username */
	username = malloc(strlen(pwd->pw_name)+strlen(at_host)+1);
	if(username == NULL) {
	  message(LOG_FATAL, "error allocating memory: %s\n", strerror(errno));
	}
	strcpy(username,pwd->pw_name);
	strcat(username,at_host);
      }
      else {				/* Use plain username for address */
	username = pwd->pw_name;
      }
				/* Send the message */
      mailto(username, subject, preamble, cursor->note_list);
    }
    else {
      message(LOG_VERBOSE,"User id %d not found in password file\n",cursor->uid);
    }
  }

}

static StringList *read_stringlist(char *fname)
{
  StringList *strlist=NULL;

  FILE *fp;

  if(  (fname == NULL) ||
       (fp = fopen(fname,"r")) == NULL ) {
				/* File cannot be opened: use NULL */
    if(fname != NULL) {
      message(LOG_ERROR,"cannot open file %s: %s\n",
	      fname, strerror(errno));
    }
  }
  else {
    char buf[80];
    StringList *cursor;	/*(ignore compiler warning may be used uninitialized)*/

			/* Read strings from file, store in stringlist */
    while( fgets(buf,sizeof(buf),fp) != NULL ) {
      StringList *new_strlist;
      char *s;
      int nchars = strlen(buf);
      if( buf[nchars-1] == '\n' ) {
	buf[--nchars] = '\0';	/* trim off the newline */
      }

      new_strlist= malloc(sizeof(StringList));
      s = malloc(nchars+1);

      if( new_strlist == NULL || s == NULL ) {
	message(LOG_FATAL, "error allocating memory: %s\n", strerror(errno));
      }
      strcpy(s,buf);

				/* Put the contents into the new node */
      new_strlist->string = s;
      new_strlist->next = NULL;

				/* Link the new node into list */
      if(strlist == NULL) {
	strlist = new_strlist; /* First node is special */
      }
      else {
	cursor->next = new_strlist; /* append new node at end */
      }
      cursor = new_strlist;
    }
  }

  return strlist;

}



/*  Mails a message to given user.
      The message is given two parts: a fixed preamble, then a part
      that is specific to the recipient.  Each part is provided as a
      linked list of strings.  The strings must be null-terminated but
      should not have newlines at the end.
 */

void mailto(char *username, char *subject,
	    StringList *preamble, StringList *message)
{

  int pfd[2];			/* Pipe file descriptors */
#define PIPEIN  pfd[0]
#define PIPEOUT pfd[1]

  pid_t child_pid;

  if( pipe(pfd) != 0 )		/* Open the parent -> child pipe */
    {
      perror("pipe open failed");
      exit(1);
    }


  if( (child_pid = fork()) != 0 ) { /* Start up a child process */

    if( child_pid == -1)	/* Fork failed */
      {
	perror("fork failed");
	exit(1);
      }


			/* Child process starts here */

    (void) close(PIPEOUT);		/* Close the writing end of pipe */

    if( dup2(PIPEIN,0) == -1)	/* Redirect stdin to come from pipe */
      {
	perror("dup2 failed");
	exit(1);
      }

				/* Become the mail program */

    if( execl(MAILPROG, MAILPROG, "-s", subject, username, NULL) == -1)
      {
	perror("execl failed");
	exit(1);
      }

  }			/* End of child process */



			/* Parent process starts here */

  else /* child_pid == 0*/ {
    StringList *curr;

    (void) close(PIPEIN);		/* Close reading end of pipe */

				/* Write the preamble to pipe */

    for(curr=preamble; curr != NULL; curr=curr->next)
      {
	write(PIPEOUT,curr->string,strlen(curr->string));
	write(PIPEOUT,"\n",1);
      }

				/* Write the notes to pipe */

    for(curr=message; curr != NULL; curr=curr->next)
      {
	write(PIPEOUT,curr->string,strlen(curr->string));
	write(PIPEOUT,"\n",1);
      }

    (void) close(PIPEOUT);		/* Close writing end of pipe: done */

  }			/* End of parent process */

}
