/* skulker.c -- remove junk files in a directory, but do it carefully.

     Based on tmpwatch.c by Erik Troan.  Additions for --junk and
     --binaries and --mail options by Robert Moniot.

   $Id: skulker.c,v 1.22 1999/09/21 15:47:55 moniot Exp $

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License
   as published by the Free Software Foundation; either version 2, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "config.h"		/* define the options */

#include <dirent.h>
#include <errno.h>
#if HAVE_GETOPT_LONG
#include <getopt.h>		/* Use the system version of getopt */
#else
#include "getopt.h"		/* Use the GNU version in local dir */
#endif
#include <stdio.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <utime.h>
#include <stdlib.h>
#include <ctype.h>

#include "message.h"
#include "mailto.h"


#define FLAGS_FORCE	(1 << 0)   /* remove even read-only stuff */
#define FLAGS_ALLFILES	(1 << 1)   /* normally just files & dirs are removed */
#define FLAGS_TEST	(1 << 2)   /* just test, don't remove anything */
#define FLAGS_JUNKFILES	(1 << 3)   /* restrict to special junkfile names */
#define FLAGS_EXES	(1 << 4)   /* restrict to executables */
#define FLAGS_OBJS	(1 << 5)   /* restrict to objects */
#define FLAGS_MPEGS	(1 << 6)   /* restrict to mpeg files */
#define FLAGS_DOSEXES	(1 << 7)   /* restrict to dos/win executables */
#define FLAGS_ZIPFILES	(1 << 8)   /* restrict to zip files */
#define FLAGS_COREMAGIC	(1 << 9)   /* use magic for core id */
#define FLAGS_AOUTMAGIC	(1 <<10)   /* use magic for a.out id */
#define FLAGS_EXESAFEDIR  (1 <<11) /* executables safe in this dir */
#define FLAGS_OBJSAFEDIR  (1 <<12) /* objects safe in this dir */
#define FLAGS_MPEGSAFEDIR (1 <<13) /* mpegs safe in this dir */
#define FLAGS_DOSSAFEDIR  (1 <<14) /* DOS EXEs safe in this dir */
#define FLAGS_ARCSAFEDIR  (1 <<15) /* Archive files safe in this dir */

				/* flags for cleaning user dirs */
#define USERCLEAN_FLAGS (FLAGS_JUNKFILES|FLAGS_EXES|FLAGS_OBJS|FLAGS_MPEGS|FLAGS_DOSEXES|FLAGS_ZIPFILES)

#define S_IPERMS (S_IRWXU | S_IRWXG | S_IRWXO) /* all permission bits */

			/* Flags for --options having no -c counterpart */
#define GETOPT_TEST	 1000
#define GETOPT_VERSION	 1001
#define GETOPT_HELP	 1002
#define GETOPT_DOSEXE	 1003
#define GETOPT_MPEG	 1004
#define GETOPT_COREMAGIC 1005
#define GETOPT_AOUTMAGIC 1006
#define GETOPT_EXE	 1007
#define GETOPT_OBJ	 1008

#ifndef MIN_UID_TO_NOTIFY
#define MIN_UID_TO_NOTIFY 100	/* don't send mail to users with uid < this */
#endif

int mail_notice = 0;	/* flag to control whether to inform the victims */

long total_bytes_cleaned;	/* Count of space freed by deleted victims */


typedef enum {			/* Define magic number categories */
  magic_unknown,
  magic_core,			/* core dump */
  magic_aout,			/* executable */
  magic_obj,			/* relocatable object */
  magic_mpeg,			/* MPEG streaming data */
  magic_dosexe,			/* DOS or Windows executable */
  magic_zip,			/* Zip archive */
} MagicNum;


				/* Define "safe" directory names where
				   various things are protected.  To make
				   a dir safe for more than one thing, OR
				   their flags.
				*/
typedef struct {
  char *dirname;
  int  flag;
} SafeDirList;


SafeDirList default_safedirs[] = {
  {"bin",	  FLAGS_EXESAFEDIR},
  {"sbin",	  FLAGS_EXESAFEDIR},
  {"cgi-bin",	  FLAGS_EXESAFEDIR},
  {"libexec",	  FLAGS_EXESAFEDIR},
  {"lib",	  FLAGS_EXESAFEDIR | FLAGS_OBJSAFEDIR},
  {"public_html", FLAGS_MPEGSAFEDIR|FLAGS_ARCSAFEDIR},
  {"archive",	  FLAGS_ARCSAFEDIR},
};

			/* Initialize the safedir list to default values. */
SafeDirList *safedirs = default_safedirs;
int num_safedirs= (sizeof(default_safedirs)/sizeof(default_safedirs[0]));
int max_safedirs= (sizeof(default_safedirs)/sizeof(default_safedirs[0]));

				/* Define list of keywords for safety classes
				   for use by add_safedir.  These simply
				   associate a keyword with the corresponding
				   flag.
				 */
SafeDirList safety_class[] = {
  {"exe",FLAGS_EXESAFEDIR},
  {"obj",FLAGS_OBJSAFEDIR},
  {"mpeg",FLAGS_MPEGSAFEDIR},
  {"dos",FLAGS_DOSSAFEDIR},
  {"arc",FLAGS_ARCSAFEDIR},
};
#define NUM_SAFETY_CLASSES (sizeof(safety_class)/sizeof(safety_class[0]))


				/* Prototypes */

int userclean(char *filename, struct stat *sb, int flags);
MagicNum get_magic_num(char *filename, struct stat *sb);
int safedir(char *filename);
int safepath(char *path);
void cleanupDirectory(char * dirname, unsigned int killTime, int flags);
void printCopyright(void);
void printUsage(void);
void printVersion(int verbose);
void printExeFlavors(FILE *fd);
void printCoreFlavors(FILE *fd);
int read_conf_file(char *conf_file, int nondefault);
int add_safedir(char *dir,char *class);

   /* malloc practically never fails, but we need to be robust */
#define malloc_or_die(PTR,SIZE) { \
  PTR=malloc(SIZE); \
  if((PTR) == NULL) { \
     message(LOG_FATAL, "error allocating memory: %s\n", strerror(errno)); \
  } }

				/* Function definitions start here */

void cleanupDirectory(char * dirname, unsigned int killTime, int flags) {
    DIR * dir;
    struct dirent * ent;
    struct stat sb;
    struct stat here;
    char *pathname;

    message(LOG_DEBUG, "cleaning up directory %s\n", dirname);

    if(flags&FLAGS_EXESAFEDIR)
      message(LOG_DEBUG,"Directory %s is safe for executables\n",dirname);
    if(flags&FLAGS_OBJSAFEDIR)
      message(LOG_DEBUG,"Directory %s is safe for objects\n",dirname);
    if(flags&FLAGS_MPEGSAFEDIR)
      message(LOG_DEBUG,"Directory %s is safe for mpegs\n",dirname);
    if(flags&FLAGS_DOSSAFEDIR)
      message(LOG_DEBUG,"Directory %s is safe for DOS/Win EXEs\n",dirname);
    if(flags&FLAGS_ARCSAFEDIR)
      message(LOG_DEBUG,"Directory %s is safe for archives\n",dirname);

    dir = opendir(dirname);

    if (lstat(dirname, &here)) {
      message(LOG_ERROR, "error statting directory %s: %s",
	      dirname, strerror(errno));
      exit(1);
    }

    if (!dir) {
      message(LOG_ERROR, "error opening directory %s: %s\n", dirname,
	      strerror(errno));
      exit(1);
    }

    do {
      errno = 0;
      ent = readdir(dir);
      if (errno) {
	message(LOG_ERROR, "error reading directory entry: %s\n", 
		strerror(errno));
	exit(1);
      }
      if (!ent) break;

      if ((ent->d_name[0] == '.' && (ent->d_name[1] == '\0' || 
	 ((ent->d_name[1] == '.') && (ent->d_name[2] == '\0'))))) 
	continue;

      message(LOG_REALDEBUG, "found directory entry %s\n", ent->d_name);

				/* Construct full path name to current file */
      if( (pathname=malloc(strlen(dirname)+strlen(ent->d_name)+2))
	  == NULL ) {
	message(LOG_ERROR, "can't malloc pathname buffer space\n");
      }
      else {
	    strcpy(pathname,dirname);
	    strcat(pathname,"/");
	    strcat(pathname,ent->d_name);

	    if (lstat(pathname, &sb)) {
		message(LOG_ERROR, "failed to lstat %s: %s\n",
			pathname, strerror(errno));
		continue;
	    }

	    if (!sb.st_uid && !(flags & FLAGS_FORCE) && 
			!(sb.st_mode & S_IWUSR)) {
		message(LOG_DEBUG, "non-writeable file owned by root "
			"skipped: %s\n", pathname);;
		continue;
	    } else if (sb.st_dev != here.st_dev) {
		message(LOG_VERBOSE, "file on different device skipped: %s\n",
			pathname);
		continue;
	    } else {

		if (S_ISDIR(sb.st_mode)) { /* sub-directory */

		  cleanupDirectory(pathname, killTime, flags|safedir(ent->d_name));

		  if( ! (flags & USERCLEAN_FLAGS) ) {
		    if (sb.st_mtime >= killTime) continue;

		    message(LOG_VERBOSE, "rmdir %s\n", pathname);

		    if (!(flags & FLAGS_TEST)) {
		      if (rmdir(pathname)) {
				/* It is not an error to fail to rmdir, since
				   dir may not be empty.  Report only if debug.
				 */
			message(LOG_DEBUG, "failed to unlink %s\n",
				pathname);
		      }
				/* Add the dir to list for notification */
		      else if( mail_notice ) {
				/* suppress notice if admin user */
			if(sb.st_uid >= MIN_UID_TO_NOTIFY) {
			   make_note(pathname,sb.st_uid);
			}
		      }
		    }
		  }
		} else {		/* non-directory */

		  if (sb.st_atime >= killTime) continue;

		  if ((flags & FLAGS_ALLFILES) || S_ISREG(sb.st_mode)) {
		    if ( !((sb.st_mode & S_IWUSR) || (flags & FLAGS_FORCE)) )
		      continue;
		    if ( flags & (USERCLEAN_FLAGS) ) {
		      if( ! userclean(pathname,&sb,flags) )
			continue;
		    }

		    message(LOG_VERBOSE, "rm %-60s\t# %7ld bytes\n", 
			    pathname,sb.st_size);

		    total_bytes_cleaned += sb.st_size;

		    if (!(flags & FLAGS_TEST)) {
		      if (unlink(pathname)) {
			message(LOG_ERROR, "failed to unlink %s\n", 
				pathname);
			total_bytes_cleaned -= sb.st_size;
		      }
				/* Add the file to list for notification */
		      else if( mail_notice ) {
				/* suppress notice if admin user */
			if(sb.st_uid >= MIN_UID_TO_NOTIFY) {
			   make_note(pathname,sb.st_uid);
			}
		      }
		    }
		  }
		}
	    }
      }
      free(pathname);
    } while (ent);

    closedir(dir);

}

				/* Returns safe flag if directory
				   name is in list of "safe" names
				 */
int safedir(char *filename)
{
  int i;

  for(i=0; i<num_safedirs; i++) {
    if(strcmp(filename,safedirs[i].dirname) == 0)
      return safedirs[i].flag;
  }

  return 0;
}

				/* Returns safe flag if any path component
				   is a safe dir.
				   NOTE: this temporarily modifies path arg.
				 */
int safepath(char *path)
{
  int is_safe=0;
  char *dir,*end;
  for(dir=path; dir != NULL && !is_safe; dir=end) {
    dir++;		/* Advance past the initial "/" */
    end = strchr(dir,'/'); /* Find the next "/" */
    if( end != NULL ) {
      *end = '\0';	/* temporarily terminate the dirname */
    }
    is_safe = safedir(dir);
    if( end != NULL ) {
      *end = '/';	/* restore the "/" */
    }
  }
  return is_safe;
}



			/* Returns 1 if file is in one of the categories
			   defined by flags & USERCLEAN_FLAGS.
			 */
int userclean(char *pathname, struct stat *sb, int flags)
{
  char *filename;	/* start of pathless filename */
  int lastpos;		/* end of filename */

  filename = strrchr(pathname,'/'); /* find start of filename */
  if( filename == NULL )
    filename = pathname;	/* can't happen, but you never know ... */
  else
    filename++;			/* otherwise move past the "/" */

  lastpos = strlen(filename)-1;	/* locate last char of filename */

			/* If --junk was given, look for filenames
			   matching any of the patterns:
				*~     (emacs backup files)
				#*#    (emacs crash recovery files)
				ed.hup (ed crash recovery files)
				a.out  (default result of compilation,
					must be owner-executable)
				core   (crash dump, must be mode 0600)
			*/
  if( flags & FLAGS_JUNKFILES ) {
    if( lastpos > 0 &&
	((filename[lastpos] == '~')
	 ||
	 (filename[0] == '#' && filename[lastpos] == '#')
	 ||
	 (strcmp(filename,"ed.hup") == 0)
	 ||
	 (strcmp(filename,"a.out") == 0 &&
	  (sb->st_mode & S_IXUSR) != 0 &&
	  (!(flags & FLAGS_AOUTMAGIC) ||
	      get_magic_num(pathname,sb) == magic_aout) )
	 ||
	 (strcmp(filename,"core") == 0 &&
	  ((sb->st_mode & S_IPERMS) == (S_IRUSR|S_IWUSR)) &&
	  (!(flags & FLAGS_COREMAGIC) ||
	      get_magic_num(pathname,sb) == magic_core))
	 ))  {
      message(LOG_REALDEBUG,
	      "file identified as junk\n");
      return 1;
    }
  }

			/* If --objects was given, look for files
			   with name *.o and having the right magic
			   number.
			 */
  if( (flags & FLAGS_OBJS) && !(flags & FLAGS_OBJSAFEDIR) ) {
    if( lastpos > 1 &&
	strcmp(filename+lastpos-1,".o") == 0 &&
	(sb->st_mode & S_IXUSR) == 0 &&
	get_magic_num(pathname,sb) == magic_obj) {
      return 1;
    }
  }

			/* If --executables was given, and not in a
			   safe directory, look for files with executable
			   permission and the right magic number.
			 */
  if( (flags & FLAGS_EXES) && !(flags & FLAGS_EXESAFEDIR) ) {
    if((sb->st_mode & S_IXUSR) != 0 &&
       get_magic_num(pathname,sb) == magic_aout) {
      return 1;
    }
  }

			/* If --dos-exes was given, look for files with
			   name *.exe (case insensitive) and the right
			   magic number.
			*/
  if( (flags & FLAGS_DOSEXES) && !(flags & FLAGS_DOSSAFEDIR) ) {
    if(lastpos > 3 &&
       strcasecmp(filename+lastpos-3, ".exe") == 0 &&
       get_magic_num(pathname,sb) == magic_dosexe) {
      return 1;
    }
  }

			/* If --mpegs was given, look for files with
			   name *.mp3 (case insensitive) and the right
			   magic number.
			 */
  if( (flags & FLAGS_MPEGS) && !(flags & FLAGS_MPEGSAFEDIR) ) {
    if(lastpos > 3 &&
       strcasecmp(filename+lastpos-3, ".mp3") == 0 &&
       
       get_magic_num(pathname,sb) == magic_mpeg) {
      return 1;
    }
  }

			/* If --zipfiles was given, look for files with
			   name *.zip (case insensitive) and the right
			   magic number.
			 */
  if( (flags & FLAGS_ZIPFILES) && !(flags & FLAGS_ARCSAFEDIR) ) {
    if(lastpos > 3 &&
       strcasecmp(filename+lastpos-3, ".zip") == 0 &&
       
       get_magic_num(pathname,sb) == magic_zip) {
      return 1;
    }
  }


			/* Nothing panned out. */

  return 0;

}

			/* Look inside  file for magic no.  If
			   we can't look, then skip it.  If we do
			   look, reset access time to what it was, in
			   case we don't end up deleting it.  */

MagicNum get_magic_num(char *pathname, struct stat *sb)
{
  int fd, nread;
  static ino_t prev_inode=0;
  static dev_t prev_dev=0;
  static MagicNum prev_magic=magic_unknown;

  unsigned char buf[132];  /* size must be big enough for max magic used */
  struct utimbuf utbuf;

			/* This routine might get called more than once
			   on the same file.  If so, don't repeat work.
			 */
  if( sb->st_ino == prev_inode && sb->st_dev == prev_dev) {
    return prev_magic;
  }

  prev_inode = sb->st_ino;
  prev_dev = sb->st_dev;

			/* Record the atime and mtime for later restoration */
  utbuf.actime = sb->st_atime;
  utbuf.modtime = sb->st_mtime;

  if( ((fd = open(pathname,O_RDONLY)) > 0) ) {
      nread = read(fd,buf,sizeof(buf));

      close(fd);		/* Clean up before proceeding */
      (void) utime(pathname,&utbuf);

				/* Now examine the cookie */
      if( nread == sizeof(buf)) {


#if BINARIES_LINUX_ELF
			/* See if it matches Linux ELF core or exe or obj */
/* magic entries used:
0	string		\177ELF		ELF
>4	byte		0		invalid class
>4	byte		1		32-bit
>4	byte		2		64-bit
>5	byte		0		invalid byte order
>5	byte		1		LSB
>>16	beshort		1		relocatable,
>>16	leshort		2		executable,
>>16	leshort		4		core file
*/
	if( (strncmp(buf,"\177ELF",4) == 0) &&
	    (buf[4] == 1 || buf[4] == 2) &&
	    (buf[5] == 1) && (buf[17] == 0) ) {

	  MagicNum magic = magic_unknown;

	  if(buf[16] == 4)
	    magic = magic_core;
	  else
	    if(buf[16] == 2)
	      magic = magic_aout;
	    else
	      if(buf[16] == 1)
		magic = magic_obj;
	  if( magic != magic_unknown ) {
	    message(LOG_REALDEBUG,
		    "file identified as Linux ELF core or exe or obj\n");
	    return (prev_magic = magic);
	  }
	}
#endif /* BINARIES_LINUX_ELF */

#if BINARIES_ALPHA_COFF
			/* See if it matches OSF core file */
/* magic entries used:
0	string		Core\001	COFF format core dump (OSF/1)
>24	string		>\0		generated by '%s'

0	string		Core\002	COFF format core dump (later versions)
>24	string		>\0		generated by '%s'
*/
	if( (strncmp(buf,"Core",4) == 0) &&
	    (buf[4] == 1 || buf[4] == 2) ) {
	  int i;
	  MagicNum magic = magic_core;
	  for(i=16; i < nread && buf[i] != '\0' ;i++) {
	    if(! isascii(buf[i]) ) {
	      message(LOG_REALDEBUG,"buf[%d]=0x%x not ascii\n",i,buf[i]);
	      magic = magic_unknown;
	      break;
	    }
	  }
	  if( magic == magic_core ) {
	    message(LOG_REALDEBUG,
		    "file identified as OSF core\n");
	    return (prev_magic = magic);
	  }
	}
#endif /* BINARIES_ALPHA_COFF */


#if BINARIES_ALPHA_COFF
			/* See if it matches Alpha COFF executable or obj */
/* magic entries used:

0	leshort		0603		COFF format alpha
>22	leshort&030000	!020000		executable
>24	leshort		0413		paged
>24	leshort		0407		object
*/
	if((buf[0] == 0203 && buf[1] == 001) &&
	   ((buf[23] & 060) != 040) && (buf[25] == 001) ) {
	  MagicNum magic = magic_unknown;
	  if(buf[24] == 013)
	    magic = magic_aout;
	  else
	    if(buf[24] == 007)
	      magic = magic_obj;
	  if( magic != magic_unknown ) {
	    message(LOG_REALDEBUG,
		    "file identified as Alpha COFF executable or obj\n");
	    return (prev_magic = magic);
	  }
	}
#endif /* BINARIES_ALPHA_COFF */


#if BINARIES_RS6K
			/* See if it matches RS/6000 executable or obj */
/* magic entries used:
0	beshort		0x01df		executable (RISC System/6000 V3.1) or obj module
*/
	if( (buf[0] == 0x01 && buf[1] == 0xdf) ) {
	  MagicNum magic = magic_unknown;
			/* Since magic number does not distinguish executable
			   from object, we fudge by looking at whether
			   its executable bit is set, or name is *.o
			 */
	  if((sb->st_mode & S_IXUSR) != 0)
	    magic = magic_aout;
	  else {
	    int p=strlen(pathname)-2;
	    if(p >=0 && strncmp(pathname+p,".o",2) == 0 )
	      magic = magic_obj;
	  }
	  if( magic != magic_unknown ) {
	    message(LOG_REALDEBUG,
		    "file identified as RS/6000 executable or obj\n");
	    return (prev_magic = magic);
	  }
	}
#endif /* BINARIES_RS6K */


			/* See if it matches MS DOS executable */
/* magic entries used:
0	string	MZ		MS-DOS executable (EXE)
>24	string	@		\b, OS/2 or MS Windows
*/
	if( (buf[0] == 'M' && buf[1] == 'Z' && buf[24] == '@')
	    ||
		/* See if it matches MS Windows executable */
/* magic entries used:
128	string		PE\0\0	MS Windows PE
*/
	    (buf[128] == 'P' && buf[129] == 'E'
	     && buf[130] == '\0' && buf[131] == '\0') ) {
	  message(LOG_REALDEBUG,
		  "file identified as MS DOS or Windows executable\n");
	  return (prev_magic = magic_dosexe);
	}


			/* See if it matches MPEG stream data */
/* magic entries used:
0	belong		0x000001b3		MPEG video stream data
0	belong		0x000001ba		MPEG system stream data
0	beshort&0xfff0	0xfff0			MPEG audio stream data
*/
	if( (buf[0] == 0x00 && buf[1] == 0x00 && buf[2] == 0x01
	     && (buf[3] == 0xb3 || buf[3] == 0xba))
	    ||
	    (buf[0] == 0xff && (buf[1]&0xf0) == 0xf0) ) {
	  message(LOG_REALDEBUG,
		  "file identified as MPEG stream data\n");
	  return (prev_magic = magic_mpeg);
	}

			/* See if it matches Zip archive file */
/* magic entries used:
0       string          PK\003\004      Zip archive data
*/
	if( buf[0] == 'P' && buf[1] == 'K' && buf[2] == 03 && buf[3] == 04 ) {
	  message(LOG_REALDEBUG,
		  "file identified as Zip archive file\n");
	  return (prev_magic = magic_zip);
	}

      }/* End if( nread...) */
  }/* End if( fd=open...) */

  message(LOG_REALDEBUG,"file magic not recognized\n");
  return (prev_magic = magic_unknown);		/* Not identified */
}



				/* Prints a list of flavors of executables
				   and objects recognized by --binaries
				 */
void printExeFlavors(FILE *fd)
{
#undef BINARIES_SOME_KIND

#if BINARIES_LINUX_ELF
#ifndef BINARIES_SOME_KIND
#define BINARIES_SOME_KIND
#endif
    fprintf(fd, "  Linux ELF");
#endif

#if BINARIES_ALPHA_COFF
#ifndef BINARIES_SOME_KIND
#define BINARIES_SOME_KIND
#endif
    fprintf(fd, "  Alpha COFF");
#endif

#if BINARIES_RS6K
#ifndef BINARIES_SOME_KIND
#define BINARIES_SOME_KIND
#endif
    fprintf(fd, "  RS/6000");
#endif
#ifndef BINARIES_SOME_KIND
    fprintf(fd, "  none");
#endif
}

			/* Prints a list of core files recognized by
			   --junk when --core-magic is given.  Not same
			   as above due to lack of magic for some kinds
			   of core file.
			 */
void printCoreFlavors(FILE *fd)
{
#undef CORE_SOME_KIND

#if BINARIES_LINUX_ELF
#ifndef CORE_SOME_KIND
#define CORE_SOME_KIND
#endif
    fprintf(fd, "  Linux ELF");
#endif

#if BINARIES_ALPHA_COFF
#ifndef CORE_SOME_KIND
#define CORE_SOME_KIND
#endif
    fprintf(fd, "  Alpha COFF");
#endif

#ifndef CORE_SOME_KIND
    fprintf(fd, "  none");
#endif
}

void printVersion(int verbose) {
  fprintf(stderr,
      "skulker version %s\n", VERSION);
  if(verbose) {
    fprintf(stderr,
      "The following types of executable and object files are recognized:\n");
    printExeFlavors(stderr);
    fprintf(stderr,"\n\n");

    fprintf(stderr,
      "The following types of core files can be verified using magic number:\n");
    printCoreFlavors(stderr);
    fprintf(stderr,"\n");
  }
  fprintf(stderr,"\nFor a usage summary use --help\n");
}

void printCopyright(void) {
    fprintf(stderr,
      "\nThis may be freely redistributed under the terms of the GNU Public License.\n");
}

void printUsage(void) {
  fprintf(stderr,
    "Usage:\n"
    "    skulker [options] <time-untouched> <dirs>\n"
    "\nOptions:\n"
    "-f,       --force              remove even if not writeable\n"
    "-a,       --all                remove all file types\n"
    "-j,       --junk               remove junk files\n"
    "          --core-magic         verify junk core files by magic number\n"
    "          --aout-magic         verify junk a.out files by magic number\n"
    "-b,       --binaries           remove executables and objects\n"
    "          --executables        remove executables\n"
    "          --objects            remove objects\n"
    "-x,       --xfer-files         remove MPEG and DOS/Win EXEs\n"
    "          --mpegs              remove MPEG files\n"
    "          --dos-exes           remove DOS/Win EXEs\n"
    "-z,       --zipfiles           remove Zip archive files\n"
    "-C fname, --config-file fname  read config file fname\n"
    "-m,       --mail               mail notices to users\n"
    "-v,       --verbose            print out what is being done\n"
    "          --test               don't delete anything\n"
    "          --help               print usage summary\n"
    "          --version            print version information\n"
    );
}

int main(int argc, char ** argv) {
    float grace;
    char graceunits;
    char *config_file = CONFIG_FILE;
    int config_file_given = 0;
    unsigned int killTime, long_index;
    int flags = 0, arg, iarg;
    struct stat sb;
    struct option options[] = {
	    { "all", 0, 0, 'a' },
	    { "aout-magic", 0, 0, GETOPT_AOUTMAGIC },
	    { "binaries", 0, 0, 'b' },
	    { "config-file", 1, 0, 'C' },
	    { "core-magic", 0, 0, GETOPT_COREMAGIC },
	    { "dos-exes", 0, 0, GETOPT_DOSEXE },
	    { "executables", 0, 0, GETOPT_EXE },
	    { "force", 0, 0, 'f' },
	    { "help", 0, 0, GETOPT_HELP },
	    { "junk", 0, 0, 'j' },
	    { "mail", 0, 0, 'm' },
	    { "mpegs", 0, 0, GETOPT_MPEG },
	    { "objects", 0, 0, GETOPT_OBJ },
	    { "test", 0, 0, GETOPT_TEST },
	    { "verbose", 0, 0, 'v' },
	    { "version", 0, 0, GETOPT_VERSION },
	    { "xfer-files", 0, 0, 'x' },
	    { "zipfiles", 0, 0, 'z' },
	    { NULL, 0,0,0 },
    };

    if (argc == 1) {
      printVersion(0);
      exit(1);
    }

    while (1) {
	long_index = 0;

	arg = getopt_long(argc, argv, "C:abfjmvxz", options, 
			  &long_index);
	if (arg == -1) break;

	switch (arg) {
	  case 'C':
	    config_file = optarg;
	    config_file_given = 1;
	    break;
	  case 'a':
	    flags |= FLAGS_ALLFILES;
	    break;

	  case 'j':
	    flags |= FLAGS_JUNKFILES;
	    break;

	  case 'b':
	    flags |= (FLAGS_EXES | FLAGS_OBJS); /* binaries */
	    break;

	  case GETOPT_EXE:
	    flags |= FLAGS_EXES;
	    break;

	  case GETOPT_OBJ:
	    flags |= FLAGS_OBJS;
	    break;

	  case 'x':
	    flags |= (FLAGS_MPEGS | FLAGS_DOSEXES); /* xfer files */
	    break;

	  case GETOPT_MPEG:
	    flags |= FLAGS_MPEGS;
	    break;

	  case GETOPT_DOSEXE:
	    flags |= FLAGS_DOSEXES;
	    break;

	  case GETOPT_COREMAGIC:
	    flags |= FLAGS_COREMAGIC;
	    break;

	  case GETOPT_AOUTMAGIC:
	    flags |= FLAGS_AOUTMAGIC;
	    break;

	  case 'f':
	    flags |= FLAGS_FORCE;
	    break;

	  case 'm':
	    mail_notice = 1;
	    break;

	  case 'v':
	    if(logLevel > 0) logLevel--;
	    break;

	  case 'z':
	    flags |= FLAGS_ZIPFILES;
	    break;

	  case GETOPT_TEST:
	    flags |= FLAGS_TEST;
	    break;

	  case GETOPT_HELP:
	    printUsage();
	    exit(0);
	    break;

	  case GETOPT_VERSION:
	    printVersion(1);
	    printCopyright();
	    exit(0);
	    break;

	  case '?':
	    fprintf(stderr,"\nFor a usage summary use --help\n");
	    exit(1);
	}
    }

    if (optind == argc) {
	fprintf(stderr, "error: time must be given\n");
	exit(1);
    }

    if( (flags & (FLAGS_COREMAGIC|FLAGS_AOUTMAGIC) &&
	 !(flags & FLAGS_JUNKFILES)) ) {
      fprintf(stderr, "error: --%s-magic specified without --junk\n",
	      (flags & FLAGS_COREMAGIC)? "core": "aout");
      exit(1);
    }

    graceunits = 'd';		/* Default units of grace period */
    if (sscanf(argv[optind], "%f%c", &grace,&graceunits) < 1) {
	fprintf(stderr, "error: bad time argument %s\n", argv[optind]);
	exit(1);
    }

				/* Convert grace period to seconds.  The
				   missing break's are intentional.  */
    switch(graceunits) {
    case 'w':
      grace *= 7.0;			/* from weeks to days */
    case 'd':
      grace *= 24.0;			/* from days to hours */
    case 'h':
      grace *= 60.0;			/* from hours to minutes */
    case 'm':
      grace *= 60.0;			/* from minutes to seconds */
    case 's':
      break;				/* from seconds to seconds */
    default:
	fprintf(stderr, "error: bad time units \"%c\"\n",graceunits);
	exit(1);
      break;
    }

    message(LOG_DEBUG, "grace period is %.0f seconds\n", grace);

				/* Read the skulker.conf file */
    if( read_conf_file(config_file, config_file_given) ) {
      message(LOG_FATAL,"Error reading config file %s\n",config_file);
    }
    message(LOG_DEBUG,"mail sent to users%s\n",at_host?at_host:" on local host");

    killTime = time(NULL) - (int)(grace+0.5);

    optind++;
    if (optind == argc) {
	fprintf(stderr, "error: directory name(s) expected\n");
	exit(1);
    }

				/* Enforce fully qualified path names */
    for(iarg=optind; iarg<argc; iarg++) {
      if( argv[iarg][0] != '/' ) {
	fprintf(stderr, "error: directory %s: path must begin with /\n",
		argv[iarg]);
	exit(1);
      }
    }

    total_bytes_cleaned = 0;

    while (optind < argc) {
	if (lstat(argv[optind], &sb)) {
	    message(LOG_ERROR, "lstat() of directory %s failed: %s\n",
		    argv[optind], strerror(errno));
	    exit(1);
	}

	if (S_ISLNK(sb.st_mode)) {
	    message(LOG_DEBUG, "initial directory %s is a symlink -- "
			"skipping\n", argv[optind]);
	} else {
	    cleanupDirectory(argv[optind], killTime, flags|safepath(argv[optind]));
	}

	optind++;
    }

    message(LOG_VERBOSE,"# %ld bytes cleaned\n",total_bytes_cleaned);


				/* Inform the victims what happened */
    if(mail_notice)
      mail_notices("Files deleted in directory cleanup", MAIL_PREAMBLE);

    return 0;
}


			/* This routine reads the configuration file.  It
			   returns nonzero if there is any error.  However,
			   being unable to open the file (e.g. does not
			   exist) is not an error (unless the user specified
			   a config file on the command line) since config
			   file is optional.
			 */

#define SPACE	" \t\n"		/* for strtok calls */

int read_conf_file(char *conf_file, int nondefault)
{
  FILE *conf_fd;
  char buf[CONFBUFSZ];		/* buffer to hold a config line */
  char *key, *name, *class;	/* tokens parsed from a line */
  int line_num=0;

  conf_fd = fopen(conf_file,"r");

  if(conf_fd == NULL) {  /* File not found is OK if default, but say why. */
    int level= nondefault? LOG_ERROR: LOG_DEBUG;
    message(level,"Cannot open config file %s: %s\n",
	    conf_file,strerror(errno));
    if( ! nondefault )
      message(LOG_DEBUG,"   -->Using defaults.\n");
    return nondefault;
  }
  message(LOG_DEBUG,"Reading config file %s\n",conf_file);

  while( fgets(buf,CONFBUFSZ,conf_fd) != NULL ) {
    ++line_num;
				/* If last char is not newline, then it means
				   we ran out of buffer size.
				 */
    if( buf[strlen(buf)-1] != '\n' ) {
      message(LOG_ERROR,"line %d exceeds max length %d in file %s\n",
	      line_num, CONFBUFSZ-1, conf_file);
      fclose(conf_fd);
      return 1;
    }
				/* Deal with comments */
    if( (key=strchr(buf,'#')) != NULL ) {
      *key = '\0';
    }

				/* Use the evil strtok function for sake
				   of portability.  Anyway, we don't care
				   if buf is clobbered.
				 */
    key = strtok(buf,SPACE);
    if(key != NULL) {

				/* Look for a safedir definition line */
      if(strcmp(key,"safedir") == 0) {
				/* Make sure there is a directory name */
	name = strtok(NULL,SPACE);
	if( name == NULL ) {
	  message(LOG_ERROR, "need directory name on line %d in file %s\n",
		  line_num, conf_file);
	  fclose(conf_fd);
	  return 1;
	}
				/* Make sure there is at least one class,
				   unless this is the special case "none".
				 */
	class=strtok(NULL,SPACE);
	if( class == NULL ) {
	  if( strcmp(name,"none") == 0 ) {
	    safedirs = NULL;	/* clear out the table */
	    num_safedirs = 0;
	  }
	  else {
	    message(LOG_ERROR, "need safedir class on line %d in file %s\n",
		    line_num, conf_file);
	    fclose(conf_fd);
	    return 1;
	  }
	}
	else {			/* Class names specified: add them to list */
	  do {
	    if( add_safedir(name,class) ) {
	      fclose(conf_fd);
	      return 1;
	    }
	  } while( (class=strtok(NULL,SPACE)) != NULL );
	}
      }	/* end of safedir definition */


				/* Look for a mailto definition line */
      else if(strcmp(key,"mailto") == 0) {
	name = strtok(NULL,SPACE);
	if( name == NULL ) {
	  message(LOG_ERROR, "need @host.domain on line %d in file %s\n",
		  line_num, conf_file);
	  fclose(conf_fd);
	  return 1;
	}
	malloc_or_die(at_host, (strlen(name)+1));
	strcpy(at_host,name);
      }	/* end of mailto definition */

      else {
	message(LOG_ERROR, "unrecognized keyword '%s' on line %d in file %s\n",
		  key, line_num, conf_file);
	fclose(conf_fd);
	return 1;
      }

    }
  }

  fclose(conf_fd);
  
  return 0;			/* success */
}

			/* This routine adds a flag for safety of a class of
			   files to a given directory in safedirlist.
			 */
int add_safedir(char *dir,char *class)
{
  int i;
  int flag=0;
  
			/* Look up the class in table to convert to flag */
  for(i=0; i<NUM_SAFETY_CLASSES; i++) {
    if(strcmp(safety_class[i].dirname,class) == 0) {
      flag = safety_class[i].flag;
      break;
    }
  }
  if( flag == 0 ) {
    message(LOG_ERROR,"No such safedir class: %s\n",class);
    return 1;
  }

				/* Now look up the dir in safedir table and
				   merge the flag if it is there already.
				 */
  for(i=0; i<num_safedirs; i++) {
    if(strcmp(safedirs[i].dirname,dir) == 0) {
      safedirs[i].flag |= flag;
      return 0;
    }
  }

				/* Dir not found: add it to the table.
				   First get more room if needed.
				 */
  if( safedirs == NULL || num_safedirs == max_safedirs ) {
    max_safedirs *= 2;		/* double the size of table */

				/* If we are still using the default table,
				   need to alloc brand-new space.
				 */
    if( safedirs == default_safedirs ) {
      int i;
      malloc_or_die(safedirs, max_safedirs*sizeof(SafeDirList));
      for(i=0; i < num_safedirs; i++) {
	safedirs[i] = default_safedirs[i];
      }
    }
    else {
      safedirs = realloc(safedirs, max_safedirs*sizeof(SafeDirList));
      if(safedirs == NULL) {
	message(LOG_FATAL,"Cannot allocate memory for safedirs: %s",
		strerror(errno));
      }
    }
  }

  {
    char *dirname;		/* Allocate permanent space for dir name */
    malloc_or_die(dirname, strlen(dir)+1);
    strcpy(dirname,dir);
    safedirs[num_safedirs].dirname = dirname;
    safedirs[num_safedirs].flag = flag;
    num_safedirs++;
  }

  return 0;			/* success */
}

#if 0				/* for debugging */
void print_safedirs(void)
{
  int i;
  for(i=0; i<num_safedirs; i++) {
    int flags = safedirs[i].flag;
    char *dirname = safedirs[i].dirname;
    if(flags&FLAGS_EXESAFEDIR)
      message(LOG_DEBUG,"Directory %s is safe for executables\n",dirname);
    if(flags&FLAGS_OBJSAFEDIR)
      message(LOG_DEBUG,"Directory %s is safe for objects\n",dirname);
    if(flags&FLAGS_MPEGSAFEDIR)
      message(LOG_DEBUG,"Directory %s is safe for mpegs\n",dirname);
    if(flags&FLAGS_DOSSAFEDIR)
      message(LOG_DEBUG,"Directory %s is safe for DOS/Win EXEs\n",dirname);
    if(flags&FLAGS_ARCSAFEDIR)
      message(LOG_DEBUG,"Directory %s is safe for archives\n",dirname);
  }
}
#endif /* 0 */
